
program Pl0;
(* PL0 COMPILER WITH CODE GENERATION : NIKLAUS WIRTH *)

 const
  HT = 9;
  Norw = 14; (* # RESERVED WORDS *)
  Txmax = 100; (* LENGTH IDENTIFIER TABLE *)
  Nmax = 14; (* MAX NUMBER OF DIGITS IN NUMBERS *)
  Al = 10; (* LENGTH OF IDENTIFIERS *)
  Amax = 2047; (* MAX ADDRESS *)
  Levmax = 3; (* MAX DEPTH OF BLOCK NESTING *)
  Cxmax = 2047; (* SIZE OF CODE ARRAY *)


 type
  Symbol = (Nul, Ident, Number, Plus, Minus, Times, Slash, Oddsym, Orsym,
            Andsym, Notsym, Eql, Neq, Lss, Leq, Gtr, Geq, Lparen, Rparen,
            Comma, Semicolon, Period, Becomes, Beginsym, Endsym, Ifsym,
            Thensym, Whilesym, Dosym, Callsym, Constsym, Varsym, Procsym);
  Alfa = packed array[1..Al] of char;
  Obj = (Constant, Variable, Proced);
  Symset = set of Symbol;
  Fct = (Lit, Opr, Lod, Sto, Cal, Int, Jmp, Jpc);
  Instruction = packed record
    F: Fct; (* FUNCTION CODE *)
    L: 0..Levmax; (*LEVEL*)
    A: 0..Amax; (* DISPLACEMENT ADDR *)
   end;


 var
  Source: text;
  Ch: char;
  Err: integer; (* ERROR COUNT *)
  Sym: Symbol;
  Id: Alfa;
  Num: integer;
  Cc: integer;
  Ll: integer;
  Kk: integer;
  Cx: integer;
  Line: array[1..255] of char;
  Temp: Alfa;
  Code: array[0..Cxmax] of Instruction;
  Word: array[1..Norw] of Alfa;
  Wsym: array[1..Norw] of Symbol;
  Ssym: array[char] of Symbol;
  Mnemonic: array[Fct] of packed array[1..3] of char;
  Declbegsys, Statbegsys, Facbegsys: Symset;
  Table: array[0..Txmax] of record
    Name: Alfa;
    case Kind : Obj of
     Constant: (
       Val: integer
     );
     Variable, Proced: (
       Level, Adr: integer
     )
   end;

 procedure Error (N: integer);
 begin
  writeln(output, ' ***', ' ' : Cc - 1, '^', N : 2);
  Err := Err + 1
 end; (*ERROR*)

 procedure Getsym;
  var
   I, J, K: integer;

  procedure Getch;
  begin
   if Cc = Ll then
    begin
     if eof(Source) then
      begin
       write(output, ' PROGRAM INCOMPLETE');
       Halt
      end;
     Ll := 0;
     Cc := 0;
     write(output, Cx : 5, ' ');
     while not eoln(Source) do
      begin
       Ll := Ll + 1;
       read(Source, Ch);
       write(output, Ch);
       Line[Ll] := Ch
      end;
     writeln;
     Ll := Ll + 1;
     readln(Source);
     Line[Ll] := ' '
    end;
   Cc := Cc + 1;
   Ch := Line[Cc]
  end; (*GETCH*)

 begin (*GETSYM*)
  while Ch in [' ', chr(HT)] do
   Getch;
  if Ch in ['A'..'Z', 'a'..'z'] then
   begin (*IDENTIFIER OR RESERVED WORD*)
    if Ch in ['a'..'z'] then
     Ch := chr(ord(Ch) - ord('a') + ord('A')); {make upper case}
    K := 0;
    repeat
     if K < Al then
      begin
       K := K + 1;
       Temp[K] := Ch
      end;
     Getch;
     if Ch in ['a'..'z'] then
      Ch := chr(ord(Ch) - ord('a') + ord('A')); {make upper case}
    until not (Ch in ['A'..'Z', '0'..'9']);
    if K >= Kk then
     Kk := K
    else
     repeat
      Temp[Kk] := ' ';
      Kk := Kk - 1
     until Kk = K;
    Id := Temp;
    I := 1;
    J := Norw;
    repeat
     K := (I + J) div 2;
     if Id <= Word[K] then
      J := K - 1;
     if Id >= Word[K] then
      I := K + 1
    until I > J;
    if I - 1 > J then
     Sym := Wsym[K]
    else
     Sym := Ident
   end
  else if Ch in ['0'..'9'] then
   begin (*NUMBER*)
    K := 0;
    Num := 0;
    Sym := Number;
    repeat
     Num := 10 * Num + (ord(Ch) - ord('0'));
     K := K + 1;
     Getch
    until not (Ch in ['0'..'9']);
    if K > Nmax then
     Error(30)
   end
  else if Ch = ':' then
   begin
    Getch;
    if Ch = '=' then
     begin
      Sym := Becomes;
      Getch
     end
    else
     Sym := Nul;
   end
  else
   begin
    Sym := Ssym[Ch];
    Getch
   end
 end; (*GETSYM*)


 procedure Gen (X: Fct; Y, Z: integer);

 begin
  if Cx > Cxmax then
   begin
    write(output, ' PROGRAM TOO LONG');
    Halt
   end;
  with Code[Cx] do
   begin
    F := X;
    L := Y;
    A := Z
   end;
  Cx := Cx + 1
 end; (*GEN*)

 procedure Test (S1, S2: Symset; N: integer);

 begin
  if not (Sym in S1) then
   begin
    Error(N);
    S1 := S1 + S2;
    while not (Sym in S1) do
     Getsym
   end
 end; (*TEST*)

 procedure Block (Lev, Tx: integer; Fsys: Symset);
  var
   Dx: integer; (*DATA ALLOCATION INDEX*)
   Tx0: integer; (*INITIAL TABLE INDEX*)
   Cx0: integer; (*INITIAL CODE INDEX*)
   Size: integer;
                    (*PROCEDURE SIZE ADDED FOR CORRECTNESS  *)

  procedure Enter (K: Obj);
  begin (*ENTER OBJECT INTO TABLE*)
   Tx := Tx + 1;
   with Table[Tx] do
    begin
     Name := Id;
     Kind := K;
     case K of
      Constant: 
       begin
       if Num > Amax then
       begin
       Error(31);
       Num := 0
       end;
       Val := Num
       end;
      Variable: 
       begin
       Level := Lev;
       Adr := Dx;
       Dx := Dx + 1;
       end;
      Proced: 
       Level := Lev
     end
    end
  end; (*ENTER*)

  function Position (Id: Alfa): integer;
   var
    I: integer;

  begin (* FIND IDENTIFIER ID IN TABLE*)
   Table[0].Name := Id;
   I := Tx;
   while Table[I].Name <> Id do
    I := I - 1;
   Position := I
  end; (*POSITION*)

  procedure Constdeclaration;
  begin
   if Sym = Ident then
    begin
     Getsym;
     if Sym in [Eql, Becomes] then
      begin
       if Sym = Becomes then
       Error(1);
       Getsym;
       if Sym = Number then
       begin
       Enter(Constant);
       Getsym
       end
       else
       Error(2)
      end
     else
      Error(3)
    end
   else
    Error(4)
  end; (*CONSTDECLARATION*)

  procedure Vardeclaration;

  begin
   if Sym = Ident then
    begin
     Enter(Variable);
     Getsym
    end
   else
    Error(4)
  end; (*VARDECLARATION*)

  procedure Listcode;
   var
    I: integer;

  begin (*LIST CODE GENERATED FOR THIS BLOCK*)
   for I := Cx0 to Cx - 1 do
    with Code[I] do
     writeln(output, I, Mnemonic[F] : 5, L : 3, A : 5)
  end; (*LISTCODE*)

  procedure Statement (Fsys: Symset);
   var
    I, Cx1, Cx2: integer;

   procedure Expression (Fsys: Symset);
    var
     Expop: Symbol;

    procedure Term (Fsys: Symset);
     var
      Termop: Symbol;

     procedure Factor (Fsys: Symset);
      var
       I: integer;
     begin
      Test(Facbegsys, Fsys, 24);
      while Sym in Facbegsys do
       begin
       if Sym = Ident then
       begin
       I := Position(Id);
       if I = 0 then
       Error(11)
       else
       with Table[I] do
       case Kind of
       Constant: 
       Gen(Lit, 0, Val);
       Variable: 
       Gen(Lod, Lev - Level, Adr);
       Proced: 
       Error(21)
       end;
       Getsym
       end
       else if Sym = Number then
       begin
       if Num > Amax then
       begin
       Error(31);
       Num := 0
       end;
       Gen(Lit, 0, Num);
       Getsym
       end
       else if Sym = Lparen then
       begin
       Getsym;
       Expression([Rparen] + Fsys);
       if Sym = Rparen then
       Getsym
       else
       Error(22)
       end;
       Test(Fsys, [Lparen], 23)
       end
     end; (*FACTOR*)

    begin (*TERM*)
     Factor(Fsys + [Times, Slash, Andsym]);
     while Sym in [Times, Slash, Andsym] do
      begin
       Termop := Sym;
       Getsym;
       Factor(Fsys + [Times, Slash, Andsym]);
       case Termop of
       Times: 
       Gen(Opr, 0, 4);
       Slash: 
       Gen(Opr, 0, 5);
       Andsym: 
       writeln('AND');
       end
      end
    end; (*TERM*)

   begin (*EXPRESSION*)
    if Sym in [Plus, Minus, Notsym] then
     begin
      Expop := Sym;
      Getsym;
      Term(Fsys + [Plus, Minus, Notsym]);
      case Expop of
       Plus: 
       ;
       Minus: 
       Gen(Opr, 0, 1);
       Notsym: 
       writeln('NOT');
      end
     end
    else
     Term(Fsys + [Plus, Minus, Orsym]);
    while Sym in [Plus, Minus, Orsym] do
     begin
      Expop := Sym;
      Getsym;
      Term(Fsys + [Plus, Minus, Orsym]);
      case Expop of
       Plus: 
       Gen(Opr, 0, 2);
       Minus: 
       Gen(Opr, 0, 3);
       Orsym: 
       writeln('OR')
      end
     end
   end; (*EXPRESSION*)

   procedure Condition (Fsys: Symset);
    var
     Relop: Symbol;
   begin
    if Sym = Oddsym then
     begin
      Getsym;
      Expression(Fsys);
      Gen(Opr, 0, 6)
     end
    else
     begin
      Expression([Eql, Neq, Lss, Gtr, Geq, Leq] + Fsys);
      if not (Sym in [Eql, Neq, Lss, Gtr, Geq, Leq]) then
       Error(20)
      else
       begin
       Relop := Sym;
       Getsym;
       Expression(Fsys);
       case Relop of
       Eql: 
       Gen(Opr, 0, 8);
       Neq: 
       Gen(Opr, 0, 9);
       Lss: 
       Gen(Opr, 0, 10);
       Geq: 
       Gen(Opr, 0, 11);
       Gtr: 
       Gen(Opr, 0, 12);
       Leq: 
       Gen(Opr, 0, 13);
       end
       end
     end
   end; (*CONDITION*)

  begin (*STATEMENT*)
   if Sym = Ident then
    begin
     I := Position(Id);
     if I = 0 then
      Error(11)
     else if Table[I].Kind <> Variable then
      begin (* ASSIGNMENT TO NON-VARIABLE *)
       Error(12);
       I := 0
      end;
     Getsym;
     if Sym = Becomes then
      Getsym
     else
      Error(13);
     Expression(Fsys);
     if I <> 0 then
      with Table[I] do
       Gen(Sto, Lev - Level, Adr)
    end
   else if Sym = Callsym then
    begin
     Getsym;
     if Sym <> Ident then
      Error(14)
     else
      begin
       I := Position(Id);
       if I = 0 then
       Error(11)
       else
       with Table[I] do
       if Kind = Proced then
       Gen(Cal, Lev - Level, Adr)
       else
       Error(15);
       Getsym
      end
    end
   else if Sym = Ifsym then
    begin
     Getsym;
     Condition([Thensym, Dosym] + Fsys);
     if Sym = Thensym then
      Getsym
     else
      Error(16);
     Cx1 := Cx;
     Gen(Jpc, 0, 0);
     Statement(Fsys);
     Code[Cx1].A := Cx
    end
   else if Sym = Beginsym then
    begin
     Getsym;
     Statement([Semicolon, Endsym] + Fsys);
     while Sym in [Semicolon] + Statbegsys do
      begin
       if Sym = Semicolon then
       Getsym
       else
       Error(10);
       Statement([Semicolon, Endsym] + Fsys)
      end;
     if Sym = Endsym then
      Getsym
     else
      Error(17)
    end
   else if Sym = Whilesym then
    begin
     Cx1 := Cx;
     Getsym;
     Condition([Dosym] + Fsys);
     Cx2 := Cx;
     Gen(Jpc, 0, 0);
     if Sym = Dosym then
      Getsym
     else
      Error(18);
     Statement(Fsys);
     Gen(Jmp, 0, Cx1);
     Code[Cx2].A := Cx
    end;
   Test(Fsys, [], 19)
  end; (*STATEMENT*)

 begin (*BLOCK*)
  Dx := 3;
  Tx0 := Tx;
  Table[Tx].Adr := Cx;
  Gen(Jmp, 0, 0);
  if Lev > Levmax then
   Error(32);
  repeat
   if Sym = Constsym then
    begin
     Getsym;
     repeat
      Constdeclaration;
      while Sym = Comma do
       begin
       Getsym;
       Constdeclaration
       end;
      if Sym = Semicolon then
       Getsym
      else
       Error(5)
     until Sym <> Ident
    end;
   if Sym = Varsym then
    begin
     Getsym;
     repeat
      Vardeclaration;
      while Sym = Comma do
       begin
       Getsym;
       Vardeclaration
       end;
      if Sym = Semicolon then
       Getsym
      else
       Error(5)
     until Sym <> Ident;
    end;
   while Sym = Procsym do
    begin
     Getsym;
     if Sym = Ident then
      begin
       Enter(Proced);
       Getsym
      end
     else
      Error(4);
     if Sym = Semicolon then
      Getsym
     else
      Error(5);
     Block(Lev + 1, Tx, [Semicolon] + Fsys);
     if Sym = Semicolon then
      begin
       Getsym;
       Test(Statbegsys + [Ident, Procsym], Fsys, 6)
      end
     else
      Error(5)
    end;
   Test(Statbegsys + [Ident], Declbegsys, 7)
  until not (Sym in Declbegsys);
  Code[Table[Tx0].Adr].A := Cx;
  with Table[Tx0] do
   begin
    Adr := Cx; (*START ADDR OF CODE *)
    Size := Dx; (*SIZE OF DATA SEGMENT*)
   end;
  Cx0 := Cx;
  Gen(Int, 0, 0);
  Statement([Semicolon, Endsym] + Fsys);
  Code[Cx0].A := Dx;
  Gen(Opr, 0, 0); (*RETURN*)
  Test(Fsys, [], 8);
  Listcode;
 end; (*BLOCK*)


 procedure Interpret;
  const
   Stacksize = 500;
  var
   P, B, T: integer; (*PROGRAM-,BASE-,TOPSTACK-REGISTERS*)
   I: Instruction; (*INSTRUCTION REGISTER*)
   S: array[1..Stacksize] of integer; (*DATASTORE*)

  function Base (L: integer): integer;
   var
    B1: integer;
  begin
   B1 := B; (*FIND BASE L LEVELS DOWN*)
   while L > 0 do
    begin
     B1 := S[B1];
     L := L - 1
    end;
   Base := B1
  end; (*BASE*)

  procedure Operator;
  begin
   with I do
    case A of (*OPERATOR*)
     0: 
      begin (*RETURN*)
       T := B - 1;
       P := S[T + 3];
       B := S[T + 2];
      end;
     1: 
      S[T] := -S[T];
     2: 
      begin
       T := T - 1;
       S[T] := S[T] + S[T + 1]
      end;
     3: 
      begin
       T := T - 1;
       S[T] := S[T] - S[T + 1]
      end;
     4: 
      begin
       T := T - 1;
       S[T] := S[T] * S[T + 1]
      end;
     5: 
      begin
       T := T - 1;
       S[T] := S[T] div S[T + 1]
      end;
     6: 
      S[T] := ord(odd(S[T]));
     8: 
      begin
       T := T - 1;
       S[T] := ord(S[T] = S[T + 1])
      end;
     9: 
      begin
       T := T - 1;
       S[T] := ord(S[T] <> S[T + 1])
      end;
     10: 
      begin
       T := T - 1;
       S[T] := ord(S[T] < S[T + 1])
      end;
     11: 
      begin
       T := T - 1;
       S[T] := ord(S[T] >= S[T + 1])
      end;
     12: 
      begin
       T := T - 1;
       S[T] := ord(S[T] > S[T + 1])
      end;
     13: 
      begin
       T := T - 1;
       S[T] := ord(S[T] <= S[T + 1])
      end;
    end;
  end; (* OPERATOR *)

  procedure Tracestatus;
  begin
   with I do
    begin
     write(output, ' ', P : 4, Mnemonic[F] : 4, L : 6, A : 10, B : 10, T : 10);
     if T >= 2 then
      writeln(output, S[T] : 10, S[T - 1] : 10)
     else
      writeln(output)
    end (*WITH*)
  end; (* TRACE STATUS *)

 begin (* INTERPRET *)
  writeln(output, ' START PL0');
  T := 0;
  B := 1;
  P := 0;
  S[1] := 0;
  S[2] := 0;
  S[3] := 0;
  writeln(output, ' ', 'P' : 4, 'OP' : 4, 'L/SOP' : 6, 'ADDR/LIT' : 10, 'B' : 10, 'T' : 10, 'S[T-1]' : 10);
  repeat
   I := Code[P];
   P := P + 1;
   Tracestatus;
   with I do
    case F of
     Lit: 
      begin
       T := T + 1;
       S[T] := A
      end;
     Opr: 
      Operator;
     Lod: 
      begin
       T := T + 1;
       S[T] := S[Base(L) + A]
      end;
     Sto: 
      begin
       S[Base(L) + A] := S[T];
       T := T - 1
      end;
     Cal: 
      begin (*GENERATE NEW BLOCK MARK*)
       S[T + 1] := Base(L);
       S[T + 2] := B;
       S[T + 3] := P;
       B := T + 1;
       P := A
      end;
     Int: 
      T := T + A;
     Jmp: 
      P := A;
     Jpc: 
      begin
       if S[T] = 0 then
       P := A;
       T := T - 1
      end;
    end (*CASE*)
  until P = 0;
  write(output, ' END OF PL0');
 end; (*INTERPRET*)

begin (* MAIN PROGRAM *)
 for Ch := 'A' to ';' do
  Ssym[Ch] := Nul;
 Word[1] := 'AND       ';
 Word[2] := 'BEGIN     ';
 Word[3] := 'CALL      ';
 Word[4] := 'CONST     ';
 Word[5] := 'DO        ';
 Word[6] := 'END       ';
 Word[7] := 'IF        ';
 Word[8] := 'NOT       ';
 Word[9] := 'ODD       ';
 Word[10] := 'OR        ';
 Word[11] := 'PROCEDURE ';
 Word[12] := 'THEN      ';
 Word[13] := 'VAR       ';
 Word[14] := 'WHILE     ';
 Wsym[1] := Andsym;
 Wsym[2] := Beginsym;
 Wsym[3] := Callsym;
 Wsym[4] := Constsym;
 Wsym[5] := Dosym;
 Wsym[6] := Endsym;
 Wsym[7] := Ifsym;
 Wsym[8] := Notsym;
 Wsym[9] := Oddsym;
 Wsym[10] := OrSym;
 Wsym[11] := Procsym;
 Wsym[12] := Thensym;
 Wsym[13] := Varsym;
 Wsym[14] := Whilesym;
 Ssym['+'] := Plus;
 Ssym['-'] := Minus;
 Ssym['*'] := Times;
 Ssym['/'] := Slash;
 Ssym['('] := Lparen;
 Ssym[')'] := Rparen;
 Ssym['='] := Eql;
 Ssym[','] := Comma;
 Ssym['.'] := Period;
 Ssym['#'] := Neq;
 Ssym['<'] := Lss;
 Ssym['>'] := Gtr;
 Ssym['$'] := Leq;
 Ssym['%'] := Geq; (* $ FOR <= AND % FOR >= *)
 Ssym[';'] := Semicolon;
 Mnemonic[Lit] := 'LIT';
 Mnemonic[Opr] := 'OPR';
 Mnemonic[Lod] := 'LOD';
 Mnemonic[Sto] := 'STO';
 Mnemonic[Cal] := 'CAL';
 Mnemonic[Int] := 'INT';
 Mnemonic[Jmp] := 'JMP';
 Mnemonic[Jpc] := 'JPC';
 Declbegsys := [Constsym, Varsym, Procsym];
 Statbegsys := [Beginsym, Callsym, Ifsym, Whilesym];
 Facbegsys := [Ident, Number, Lparen];
 Err := 0; (* INITIALIZE ERROR COUNT *)
 Cc := 0;
 Cx := 0;
 Ll := 0;
 Ch := ' ';
 Kk := Al;
 assign(Source, 'pl0.dat');
 reset(Source);
 Getsym;
 Block(0, 0, [Period] + Declbegsys + Statbegsys);
 if Sym <> Period then
  Error(9);
 if Err = 0 then
  Interpret
 else
  write(output, ' ERRORS IN PL0 PROGRAM');
end.
